@@ -9,11 +9,11 @@ try:
     from . import manifold_torch
     from . import utils_torch
 except ImportError:
-    print('No torch package installed')
+    print('Warning: no torch package installed')
 
 
 try:
     import jax 
     from .core  import backbone_jax
 except ImportError:
-    print('No JAX package installed')
+    print('Warning: no JAX package installed')
