@@ -6,7 +6,7 @@ from .basic_manifold_torch import basic_manifold_torch
 
 
 class generalized_stiefel_torch(basic_manifold_torch):
-    def __init__(self, var_shape, B = None, device = torch.device('cpu'), dtype = torch.float64) -> None:
+    def __init__(self, var_shape, B = lambda X:X , device = torch.device('cpu'), dtype = torch.float64) -> None:
         n = var_shape[0]
         p = var_shape[1]
         self._n = n
@@ -15,10 +15,7 @@ class generalized_stiefel_torch(basic_manifold_torch):
         self.var_shape = (n,p)
         self.manifold_type = 'S'
 
-        if B is None:
-            self.B = lambda X:X 
-        else: 
-            self.B = B 
+        self.B = B 
         # Here we require B be a function since it could fully utilize the
         # low-rank structure of B
         #  when one wishes to input a matrix B_mat
