@@ -6,16 +6,21 @@ from .basic_manifold_torch import basic_manifold_torch
 
 class stiefel_torch(basic_manifold_torch):
     def __init__(self, var_shape, device = torch.device('cpu'), dtype = torch.float64) -> None:
-        self._n = var_shape[0]
-        self._p = var_shape[1]
+        if len(var_shape) >= 2:
+            self._n = var_shape[-2]
+            self._p = var_shape[-1]
+            self.dim = var_shape[:-2]
+        else:
+            print("The length of var_shape should be no less than 2.")
+            raise TypeError
 
         self.device = device
         self.dtype = dtype
 
-        super().__init__('Stiefel',(self._n,self._p), (self._p,self._p),   device= self.device ,dtype= self.dtype)
+        super().__init__('Stiefel',var_shape, (*self.dim, self._p,self._p),   device= self.device ,dtype= self.dtype)
         # self._parameters = OrderedDict()
-        self._parameters['In'] = torch.eye(self._p).to(device = self.device, dtype = self.dtype)
-        self.Ip = self._parameters['In']
+        self._parameters['Ip'] = torch.diag_embed(torch.ones((*self.dim, self._p), device=self.device, dtype=self.dtype))
+        self.Ip = self._parameters['Ip']
 
 
         
@@ -23,63 +28,65 @@ class stiefel_torch(basic_manifold_torch):
 
 
     def Phi(self, M):
-        return (M + M.T )/2
+        return (M + M.transpose(-2,-1) )/2
 
 
 
     def A(self, X):
-        XX = X.T @ X
-        return 1.5 * X - X @ (XX /2)
+        XX = torch.matmul(X.transpose(-2,-1), X)
+        return 1.5 * X - torch.matmul(X , (XX /2))
 
 
     def JA(self, X, G):
-        return G @ ( self.Ip - 0.5 * self.C(X) )  - X @ self.Phi(X.T @ G)
+        return torch.matmul(G , ( self.Ip - 0.5 * self.C(X) ))  - torch.matmul(X , self.Phi(torch.matmul(X.transpose(-2,-1), G)) )
 
     def JA_transpose(self,X,G):
         # JA is self-adjoint
         return self.JA(X,G)
 
     def hessA(self, X, gradf, D):
-        return - D @ self.Phi( X.T @ gradf  ) - X @ self.Phi(D.T @ gradf) - gradf @ self.Phi( D.T @ X )
+        return - torch.matmul(D , self.Phi( torch.matmul(X.transpose(-2,-1), gradf)  )) - torch.matmul(X , self.Phi( torch.matmul(D.transpose(-2,-1) , gradf)  )) - torch.matmul(gradf , self.Phi( torch.matmul(D.transpose(-2,-1) , X) ))
 
 
     def JC(self, X, Lambda):
-        return 2*X @ self.Phi(Lambda)
+        return torch.matmul(2*X , self.Phi(Lambda))
 
     
     def C(self, X):
         # return X.T @ X - self.Ip.to(device=X.device, dtype = X.dtype)
-        return X.T @ X - self.Ip
+        return torch.matmul(X.transpose(-2,-1), X) - self.Ip
 
     def C_quad_penalty(self, X):
         return torch.sum(self.C(X) ** 2)
 
 
     def hess_feas(self, X, D):
-        return 4*X @ self.Phi( X.T @ D ) + 2*D @ self.C(X)
+        return torch.matmul(4*X , self.Phi( torch.matmul(X.transpose(-2,-1), D) )  ) + 2* torch.matmul(D , self.C(X))
 
     
 
 
 
     def Feas_eval(self, X):
-        return torch.linalg.norm( self.C(X) , 'fro')
+        return torch.linalg.norm( self.C(X).flatten() , 2)
 
     def Init_point(self, Xinit = None):
         if Xinit is None:
-            Xinit = torch.randn(self._n, self._p).to(device = self.device, dtype = self.dtype)
+            Xinit = torch.randn(*self.var_shape).to(device = self.device, dtype = self.dtype)
         else:
             Xinit = Xinit.detach()
         Xinit = Xinit.to(device = self.device, dtype = self.dtype)
 
+        if self.Feas_eval(Xinit) > 1e-6:
+            UX, SX, VX = torch.svd(Xinit)
+            Xinit = torch.matmul(UX, VX.transpose(-2,-1))
+        
         Xinit.requires_grad = True
-        if torch.linalg.norm(Xinit.T @ Xinit - self.Ip, 'fro') > 1e-6:
-            Xinit, Rinit = torch.linalg.qr(Xinit)
         return Xinit
 
     def Post_process(self,X):
-        UX, SX, VX = torch.linalg.svd(X, full_matrices = False)
-        return UX @ VX
+        UX, SX, VX = torch.svd(X)
+        return torch.matmul(UX, VX.transpose(-2,-1))
 
 
 
@@ -96,88 +103,34 @@ class stiefel_torch(basic_manifold_torch):
         return local_obj_fun  
 
 
-    # def to_cdf_fun(self, beta = 0):
-    #     def decorator_cdf_obj(obj_fun):
-    #         return self.generate_cdf_fun(obj_fun, beta )
-    #         # return lambda X: obj_fun(self.A(X)) + (beta) * self.C_quad_penalty(X)
-            
-    #     return decorator_cdf_obj
-
 
 
     def generate_cdf_grad(self, obj_grad, beta):
         def local_grad(X):
             CX = self.C(X)
-            AX = X - 0.5 * X@CX
+            AX = X - 0.5 * torch.matmul(X,CX)
             gradf = obj_grad(AX)
-            XG = self.Phi(X.T @ gradf)
+            XG = self.Phi( torch.matmul(X.transpose(-2,-1) , gradf) )
 
             # local_JA_gradf = gradf @ (np.eye(self._p) - 0.5 * CX) - X @ XG 
             
             # local_JC_CX = 2 * X @(CX)
 
-            return gradf @ (self.Ip - 0.5 * CX) +  X @  ( 2* beta * CX - XG)
+            return torch.matmul(gradf , (self.Ip - 0.5 * CX)) +  torch.matmul(X ,  ( 2* beta * CX - XG))
 
         return local_grad  
 
-    # def to_cdf_grad(self, beta = 0):
-    #     def decorator_cdf_grad(obj_grad):
-    #         return self.generate_cdf_grad(obj_grad, beta )
-    #         # return lambda X: obj_fun(self.A(X)) + (beta) * self.C_quad_penalty(X)
-            
-    #     return decorator_cdf_grad
-
-
-
-    #TODO 
-    # 2022/01/05
-    # The simplest expression for self.generate_cdf_hess()  is
-    # self.JA( X, hessf( self.JA_transpose(X,D) ) ) + self.hessA(X, gradf, D) + beta * self.hess_feas(X, D)
-    # However, it repeatively computes A(X), X.T @ D and C(X), leading to inferior efficiency in practice.
-    # A better implementation is presented below.
-    # However, the function self.generate_cdf_hess()  is still not well-optimized. 
-    # In the nest version I will rewrite this function for a better performance.
-
-    
-    # 2022/01/07
-    # Rewite  self.generate_cdf_grad() and self.generate_cdf_hess()
-    
-
-
-    # def generate_cdf_hess(self, obj_grad, obj_hess, beta):
-    #     def local_hess(X, D):
-    #         CX = self.C(X)
-    #         AX = X - 0.5 * X@CX
-    #         gradf = obj_grad(AX)
-    #         XG = self.Phi(X.T @ gradf)
-    #         XD = self.Phi(X.T @ D)
-
-    #         local_JAT_D = D @ (np.eye(self._p) - 0.5 * CX) - X @ XD 
-    #         local_objhess_JAT_D = obj_hess(AX, local_JAT_D)
-    #         local_JA_objhess_JAT_D = local_objhess_JAT_D @ (np.eye(self._p) - 0.5 * CX) -  X @ self.Phi( X.T @ local_objhess_JAT_D )
-
-
-    #         local_hessA_objgrad_D = - D @ XG - X @ self.Phi(D.T @ gradf) - gradf @ XD
-
-    #         local_hess_feas = 4*X @ XD + 2*D @ CX
-
-
-    #         return local_JA_objhess_JAT_D + local_hessA_objgrad_D + beta * local_hess_feas
-
-    #     return local_hess
-
-
 
 
     def generate_cdf_hess(self, obj_grad, obj_hess, beta):
         def local_hess(X, D):
             CX = self.C(X)
-            AX = X - 0.5 * X@CX
+            AX = X - 0.5 *  torch.matmul(X,CX)
             gradf = obj_grad(AX)
-            XG = self.Phi(X.T @ gradf)
-            XD = self.Phi(X.T @ D)
+            XG = self.Phi( torch.matmul(X.transpose(-2,-1) , gradf) )
+            XD = self.Phi( torch.matmul(X.transpose(-2,-1) , D) )
 
-            local_JAT_D = D @ (self.Ip - 0.5 * CX) - X @ XD 
+            local_JAT_D = torch.matmul(D , (self.Ip - 0.5 * CX)) - torch.matmul(X , XD)
             local_objhess_JAT_D = obj_hess(AX, local_JAT_D)
             # local_JA_objhess_JAT_D = local_objhess_JAT_D @ (np.eye(self._p) - 0.5 * CX) -  X @ self.Phi( X.T @ local_objhess_JAT_D )
 
@@ -186,7 +139,9 @@ class stiefel_torch(basic_manifold_torch):
             # local_hess_feas = 4*X @ XD + 2*D @ CX
             # return local_JA_objhess_JAT_D + local_hessA_objgrad_D + beta * local_hess_feas
 
-            return local_objhess_JAT_D @ (self.Ip - 0.5 * CX) -  X @ self.Phi( X.T @ local_objhess_JAT_D + self.Phi(D.T @ gradf) - 4* beta * XD) + D @ (2*beta*CX - XG) - gradf @ XD
+            return (   torch.matmul(local_objhess_JAT_D , (self.Ip - 0.5 * CX)) 
+                    -  torch.matmul(X , self.Phi( torch.matmul(X.transpose(-2,-1) , local_objhess_JAT_D) + self.Phi(torch.matmul(D.transpose(-2,-1) , gradf)) - 4* beta * XD) )
+                    + torch.matmul(D , (2*beta*CX - XG) - torch.matmul(gradf , XD)  )   )
 
 
 
@@ -197,12 +152,12 @@ class stiefel_torch(basic_manifold_torch):
     def generate_cdf_hess_approx(self, obj_grad, obj_hess, beta):
         def local_hess(X, D):
             CX = self.C(X)
-            AX = X
+            AX = X - 0.5 *  torch.matmul(X,CX)
             gradf = obj_grad(AX)
-            XG = self.Phi(X.T @ gradf)
-            XD = self.Phi(X.T @ D)
+            XG = self.Phi( torch.matmul(X.transpose(-2,-1) , gradf) )
+            XD = self.Phi( torch.matmul(X.transpose(-2,-1) , D) )
 
-            local_JAT_D = D  - X @ XD 
+            local_JAT_D = D  - torch.matmul(X , XD) 
             local_objhess_JAT_D = obj_hess(AX, local_JAT_D)
             # local_JA_objhess_JAT_D = local_objhess_JAT_D @ (np.eye(self._p) - 0.5 * CX) -  X @ self.Phi( X.T @ local_objhess_JAT_D )
 
@@ -211,8 +166,7 @@ class stiefel_torch(basic_manifold_torch):
             # local_hess_feas = 4*X @ XD + 2*D @ CX
             # return local_JA_objhess_JAT_D + local_hessA_objgrad_D + beta * local_hess_feas
 
-            return local_objhess_JAT_D  -  X @ self.Phi( X.T @ local_objhess_JAT_D + self.Phi(D.T @ gradf) - 4* beta * XD) + D @ (2*beta*CX - XG) - gradf @ XD
-
+            return local_objhess_JAT_D  -  torch.matmul(X , self.Phi( torch.matmul(X.transpose(-2,-1) , local_objhess_JAT_D) + self.Phi(torch.matmul(D.transpose(-2,-1) , gradf)) - 4* beta * XD) ) + torch.matmul(D , (2*beta*CX - XG) - torch.matmul(gradf , XD)  ) 
 
 
         return local_hess
